/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "save.h"

void deleteDirectoryContents(char *name)
{
	String filename;
	
	DIR *dir = opendir(name);
	dirent *file;
	
	while ((file = readdir(dir)))
	{
		if (file->d_name[0] == '.')
		{
			continue;
		}
		
		debug(("deleteDirectoryContents() - Removing file '%s'\n", file->d_name));
		
		filename.setText("%s/%s", name, file->d_name);
		
		remove(filename.getText());
	}
	
	closedir(dir);
}

void copyDir(char *oldDir, char *targetDir)
{
	debug(("copyDir() - Copying files in '%s' to '%s'\n", oldDir, targetDir));
	
	DIR *dir = opendir(oldDir);
	dirent *file;
	
	String targetFile;
	
	while ((file = readdir(dir)))
	{
		if (file->d_name[0] == '.')
		{
			continue;
		}
		
		engine->loadData(file->d_name);
		
		targetFile.setText("%s/%s", targetDir, file->d_name);
		
		debug(("copyDir() - Copying '%s'\n", targetFile.getText()));
		
		FILE *fp = fopen(targetFile.getText(), "wb");
		
		// oh dear... :|
		if (!fp)
		{
			printf("WARNING: copyDir() - Could not copy '%s' to '%s'!\n", file->d_name, targetFile.getText());
			continue;
		}
		
		fwrite(engine->dataBuffer, 1, engine->lastReadDataSize, fp);
				
		fclose(fp);
	}
	
	closedir(dir);
}

void reorganiseSaveFiles(int oldSaveSlot, int newSaveSlot)
{
	debug(("reorganiseSaveFiles() %d -> %d\n", oldSaveSlot, newSaveSlot));
	
	String oldDir, newDir;
	
	oldDir.setText("%s/"SAVE_FORMAT, engine->userHomeDirectory.getText(), oldSaveSlot);
	newDir.setText("%s/"SAVE_FORMAT, engine->userHomeDirectory.getText(), newSaveSlot);
	
	if (oldSaveSlot == newSaveSlot)
	{
		return;
	}
	
	debug(("Removing directory '%s'\n", newDir.getText()));
	
	// Delete all the files here...
	deleteDirectoryContents(newDir.getText());
	
	// Don't worry, we're done
	if (oldSaveSlot == -1)
	{
		return;
	}
	
	// now for the tricky one... Copy the old data files to the new directory :S
	copyDir(oldDir.getText(), newDir.getText());
}

void showSaveGameProgess(char *progress, bool autoSaving)
{
	graphics->setMode(MODE_2D);
	
	if (!autoSaving)
	{
		graphics->drawRect(0, (graphics->screen->h / 2) - 10, graphics->screen->w, 34, GLColor::black, false);
		graphics->drawRect(0, (graphics->screen->h / 2) - 10, graphics->screen->w, 1, GLColor::white, false);
		graphics->drawRect(0, (graphics->screen->h / 2) + 24, graphics->screen->w, 1, GLColor::white, false);
		
		graphics->drawString(graphics->screen->w / 2, graphics->screen->h / 2, TXT_CENTERED, GLColor::white, true, progress);
		
		graphics->delay(1, false);
	}
	else
	{
		graphics->showLoadingProgress("Saving Game State...");
	}
}

void removeSaveFromList(int slotId)
{
	debug(("removeSaveFromList() - %d\n", slotId));
	
	String directoryName;
	directoryName.setText("%s/"SAVE_FORMAT, engine->userHomeDirectory.getText(), slotId);
	
	deleteDirectoryContents(directoryName.getText());
	
	debug(("removeSaveFromList() - Removing directory '%s'\n", directoryName.getText()));
	
	remove(directoryName.getText());
}

// Ahhhhhh... nice brute for method of getting save directory ;)
int getNextSlotId()
{
	debug(("getNextSlotId()\n"));
	
	String saveDir;
	saveDir.setText("%s", engine->userHomeDirectory.getText());
	
	DIR *dir = opendir(saveDir.getText());
	dirent *file;
	
	int slot = 0;
	int rtn = -1;
	
	while ((file = readdir(dir)))
	{
		if (file->d_name[0] == '.')
		{
			continue;
		}
		
		slot = atoi(file->d_name);
		
		rtn = max(rtn, slot);
	}
	
	closedir(dir);
	
	rtn++;
	
	debug(("getNextSlotId() - Returning %d\n", rtn));
	
	return rtn;
}

void createSaveHeader(int slotId)
{
	if (slotId == -1)
	{
		slotId = getNextSlotId();
	}
	
	String dir;
	dir.setText("%s/"SAVE_FORMAT, engine->userHomeDirectory.getText(), slotId);
	
	if ((mkdir(dir.getText(), S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH) != 0) && (errno != EEXIST))
	{
		graphics->showErrorAndExit("Couldn't create save directory '%s'", dir.getText());
		exit(1);
	}
	
	reorganiseSaveFiles(game->saveSlot, slotId);
	
	String outputFilename;
	outputFilename.setText("%s/"SAVE_FORMAT"/saveinfo", engine->userHomeDirectory.getText(), slotId);
	
	char timeBuffer[40];
	char dateBuffer[40];

	time_t ltime;
	time(&ltime);
	tm *today = localtime(&ltime);

	strftime(timeBuffer, 40, "%H:%M %Z", today);
	strftime(dateBuffer, 40, "%d/%m/%Y", today);
	
	FILE *fp = fopen(outputFilename.getText(), "wb");
	
	printf("Saving to '%s'\n", outputFilename.getText());
	
	if (!fp)
	{
		graphics->showErrorAndExit("createSaveHeader() - Couldn't open file '%s'", outputFilename.getText());
	}
	
	fprintf(fp, "%s;\n", timeBuffer);
	fprintf(fp, "%s;\n", dateBuffer);
	fprintf(fp, "%s;\n", localeManager->translate("MISSION_NAME"));
	fprintf(fp, "%s;\n", game->mission.getText());
	
	fclose(fp);
	
	debug(("createSaveHeader() - Slot = %d -> %d\n", game->saveSlot, slotId));
	
	game->saveSlot = slotId;
}

void compressSaveGameData(char *inputFilename)
{
	uLongf cSize = 0;
	uLongf fSize = 0;
	Bytef *buffer;
	Bytef *output;
	
	FILE *infile = fopen(inputFilename, "rb");
	fseek(infile, SEEK_SET, SEEK_END);
	long size = ftell(infile);
	fclose(infile);
	
	size = (uLongf)(size * 1.1);
	
	buffer = new Bytef[size];
	output = new Bytef[size];
	
	gzFile fp = gzopen(inputFilename, "rb");
	fSize = gzread(fp, buffer, size);
	gzclose(fp);
	
	cSize = (uLongf)((fSize * 1.01) + 12);
	compress2(output, &cSize, buffer, fSize, 9);
	
	debug(("Attempting to create directory "SAVE_FORMAT"\n", game->saveSlot));
	
	String dir;
	dir.setText("%s/"SAVE_FORMAT, engine->userHomeDirectory.getText(), game->saveSlot);
	
	String outputFilename;
	
	outputFilename.setText("%s/%s", dir.getText(), game->mission.getText());
	
	debug(("Attempting to compress save game %s to %s\n", inputFilename, outputFilename.getText()));
	
	FILE *outfile = fopen(outputFilename.getText(), "wb");
	fwrite(output, 1, cSize, outfile);
	fwrite(&cSize, 1, sizeof(uLongf), outfile);
	fwrite(&fSize, 1, sizeof(uLongf), outfile);
	fclose(outfile);
	
	/*
	remove(inputFilename);
	*/
	
	debug(("Compressed save from %d to %d\n", (int)fSize, (int)cSize));
	
	delete buffer;
	delete output;
}

void saveComment(char *comment, FILE *fp)
{
	fwrite(comment, 1, strlen(comment), fp);
}

void saveGameData(int saveSlot, bool autoSaving)
{
	engine->startDebugTimer("saveGameData()");
	
	String filename;
	filename.setText("%s/blobWarsSaveTempFile", engine->userHomeDirectory.getText());

	debug(("Attempting to save game to %s\n", filename.getText()));
	
	FILE *fp = fopen(filename.getText(), "wb");

	if (!fp)
	{
		graphics->showErrorAndExit("Couldn't save game");
	}
	
	showSaveGameProgess("Saving Game - Please Wait...", autoSaving);
	
	if (!game->changingLevel)
	{
		graphics->delay(500, false);
	}
	
	createSaveHeader(saveSlot);
	
	showSaveGameProgess("Saving Game - Please Wait... 10%%", autoSaving);
	
	saveComment("// ############### Game Data ############\n\n", fp);
	
	entityManager->saveCustomItems(fp);
	
	saveComment("// ############### Custom Item Data ############\n\n", fp);
	
	game->save(fp);
	
	showSaveGameProgess("Saving Game - Please Wait... 25%%", autoSaving);
	
	saveComment("// ############### Mission Data ############\n\n", fp);
	
	mission->save(fp);
	
	showSaveGameProgess("Saving Game - Please Wait... 50%%", autoSaving);
	
	saveComment("// ############### All Entity Data ############\n\n", fp);
	
	entityManager->save(fp);
	
	showSaveGameProgess("Saving Game - Please Wait... 75%%", autoSaving);
	
	saveComment("// ############### Particle Data ############\n\n", fp);
	
	particleManager->save(fp);
	
	showSaveGameProgess("Saving Game - Please Wait... 85%%", autoSaving);

	fprintf(fp, "@END@\n");

	fclose(fp);
	
	showSaveGameProgess("Saving Game - Please Wait... 95%%", autoSaving);
	
	compressSaveGameData(filename.getText());
	
	debug(("Successfully saved game to %s (slot %d)\n", filename.getText(), saveSlot));
	
	showSaveGameProgess("Saving Game - Done", autoSaving);
	
	engine->endDebugTimer("saveGameData()");
	
	if (!game->changingLevel)
	{
		graphics->delay(500, false);
	}
}

bool saveGame()
{
	int saveSlot = doLoadSaveGameDialog("Save");
	
	// don't bother! They pressed cancel!
	if (saveSlot == -2)
	{
		return false;
	}
	
	uiManager->onlyShowGroup("@none@");
	
	saveGameData(saveSlot, false);
	
	engine->resetTimeDifference();
	
	return true;
}
